/*
 * Copyright (c) 2010-2024. Institut Pasteur.
 *
 * This file is part of Icy.
 * Icy is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Icy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Icy. If not, see <https://www.gnu.org/licenses/>.
 */

package plugins.adufour.quickhull;

import org.bioimageanalysis.icy.extension.plugin.abstract_.Plugin;
import org.bioimageanalysis.icy.extension.plugin.annotation_.IcyPluginName;
import org.jetbrains.annotations.NotNull;

import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.List;

@IcyPluginName("QuickHull 2D")
public class QuickHull2D extends Plugin {
    public static @NotNull List<Point2D> computeConvexEnvelope(final @NotNull List<Point2D> points) {
        final ArrayList<Point2D> envelope = new ArrayList<>();

        // find two points: right (bottom) and left (top)

        Point2D l = points.getFirst();
        Point2D r = points.getFirst();

        for (int i = 1; i < points.size(); i++) {
            final Point2D p = points.get(i);
            if (r.getX() > p.getX() || (r.getX() == p.getX() && r.getY() > p.getY())) r = p;
            if (l.getX() < p.getX() || (l.getX() == p.getX() && l.getY() < p.getY())) l = p;
        }

        // create neighbor lists

        final ArrayList<Point2D> neighbors1 = new ArrayList<>();
        final ArrayList<Point2D> neighbors2 = new ArrayList<>();

        for (final Point2D p : points) {
            if (p == l || p == r) continue;

            final int upper = Line2D.relativeCCW(r.getX(), r.getY(), l.getX(), l.getY(), p.getX(), p.getY());

            if (upper > 0) {
                neighbors1.add(p);
            }
            else if (upper < 0) {
                neighbors2.add(p);
            }
        }

        envelope.add(r);

        quickhull(envelope, r, l, neighbors1);

        envelope.add(l);

        quickhull(envelope, l, r, neighbors2);

        return envelope;
    }

    private static void quickhull(final ArrayList<Point2D> envelope, final Point2D a, final Point2D b, final @NotNull ArrayList<Point2D> neighbors) {
        if (neighbors.isEmpty()) return;

        final Point2D c = farthestpoint(a, b, neighbors);

        final ArrayList<Point2D> al1 = new ArrayList<>();
        final ArrayList<Point2D> al2 = new ArrayList<>();

        for (final Point2D pt : neighbors) {
            if (pt == a || pt == b) continue;

            if (Line2D.relativeCCW(a.getX(), a.getY(), c.getX(), c.getY(), pt.getX(), pt.getY()) > 0) {
                al1.add(pt);
            }
            else if (Line2D.relativeCCW(c.getX(), c.getY(), b.getX(), b.getY(), pt.getX(), pt.getY()) > 0) {
                al2.add(pt);
            }
        }

        quickhull(envelope, a, c, al1);

        envelope.add(c);

        quickhull(envelope, c, b, al2);
    }

    private static Point2D farthestpoint(final Point2D a, final Point2D b, final @NotNull ArrayList<Point2D> points) {
        double maxD = -1;
        Point2D maxP = null;

        for (final Point2D p : points) {
            if (p == a || p == b) continue;

            final double dist = Line2D.ptLineDistSq(a.getX(), a.getY(), b.getX(), b.getY(), p.getX(), p.getY());

            if (dist > maxD) {
                maxD = dist;
                maxP = p;
            }
        }

        return maxP;
    }
}
